"""The cimsparql.model module contains the base class CimModel."""

import asyncio
import re
from dataclasses import dataclass
from functools import cached_property
from string import Template
from typing import Dict, List, Optional

import pandas as pd

from cimsparql import templates
from cimsparql.data_models import (
    AcLinesSchema,
    BordersSchema,
    BranchComponentSchema,
    BranchWithdrawSchema,
    BusDataSchema,
    ConnectionsSchema,
    ConvertersSchema,
    CoordinatesSchema,
    DcActiveFlowSchema,
    DisconnectedSchema,
    ExchangeSchema,
    FullModelSchema,
    LoadsSchema,
    MarketDatesSchema,
    PowerFlowSchema,
    RegionsSchema,
    StationGroupCodeNameSchema,
    SubstationVoltageSchema,
    SynchronousMachinesSchema,
    TransfConToConverterSchema,
    TransformersSchema,
    TransformerWindingSchema,
    WindGeneratingUnitsSchema,
)
from cimsparql.graphdb import AsyncGraphDBClient, GraphDBClient, ServiceConfig
from cimsparql.type_mapper import TypeMapper


@dataclass
class ModelConfig:
    system_state_repo: Optional[str] = None
    ssh_graph: str = "?ssh_graph"


class Model:
    def __init__(
        self,
        client: GraphDBClient,
        config: Optional[ModelConfig] = None,
        mapper: Optional[TypeMapper] = None,
    ) -> None:
        self.client = client
        self.config = config or ModelConfig()
        self.mapper = mapper or TypeMapper(client.service_cfg)

    @staticmethod
    def _col_map(data_row) -> Dict[str, str]:
        return {
            column: data.get("datatype", data.get("type", None))
            for column, data in data_row.items()
        }

    @classmethod
    def col_map(cls, data_row, columns: Dict[str, str]) -> Dict[str, str]:
        columns = columns or {}
        col_map = cls._col_map(data_row)
        col_map.update(columns)
        return col_map

    @property
    def map_data_types(self) -> bool:
        pref = self.client.prefixes
        return "cim" in pref and self.mapper.have_cim_version(pref["cim"])

    def _convert_result(
        self,
        result: pd.DataFrame,
        data_row: List[dict],
        index: Optional[str] = None,
        columns: Optional[Dict[str, str]] = None,
    ) -> pd.DataFrame:
        col_map = self.col_map(data_row, columns)
        result = self.mapper.map_data_types(result, col_map)

        if index:
            return result.set_index(index)
        return result

    async def get_table_and_convert(
        self, query: str, index: Optional[str] = None, columns: Optional[Dict[str, str]] = None
    ) -> pd.DataFrame:
        if isinstance(self.client, AsyncGraphDBClient):
            result, data_row = await self.client.get_table(query)
        else:
            result, data_row = self.client.get_table(query)
        return self._convert_result(result, data_row, index, columns)

    @property
    def state_repo(self) -> str:
        return self.config.system_state_repo or self.client.service_cfg.url

    def template_to_query(
        self, template: Template, substitutes: Optional[Dict[str, str]] = None
    ) -> str:
        """Convert provided template to query."""
        if substitutes is None:
            substitutes = {}

        return template.safe_substitute(
            {"repo": self.state_repo} | substitutes | self.client.prefixes
        )


class CimModel(Model):
    """Used to query with sparql queries (typically CIM)."""

    @cached_property
    def cim_version(self) -> int:
        return int(re.search("cim(\\d+)", self.client.prefixes["cim"]).group(1))

    @property
    def full_model_query(self) -> str:
        return self.template_to_query(templates.FULL_MODEL_QUERY)

    async def full_model(self) -> FullModelSchema:
        """Return all models where all depencies has been created and is available

        All profiles EQ/SSH/TP/SV will define a md:FullModel with possible dependencies. One profile
        could be dependent on more than one other. This function will return the models for SSH/TP
        and SV that is available from current repo or provided <repo>.

        Example:
        >> model.full_model()
        """
        df = await self.get_table_and_convert(self.full_model_query)
        return FullModelSchema(df)

    @property
    def market_dates_query(self) -> str:
        """Market activation date for this repository."""
        return self.template_to_query(templates.MARKET_DATES_QUERY)

    async def market_dates(self) -> MarketDatesSchema:
        df = await self.get_table_and_convert(self.market_dates_query, index="mrid")
        return MarketDatesSchema(df)

    def bus_data_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.BUS_DATA_QUERY, substitutes)

    def three_winding_dummy_nodes_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.THREE_WINDING_DUMMY_NODES_QUERY, substitutes)

    async def bus_data(self, region: Optional[str] = None) -> BusDataSchema:
        """Query name of topological nodes (TP query).

        Args:
           region: Limit to region (use None to get all)
        """
        dfs = await asyncio.gather(
            self.get_table_and_convert(self.bus_data_query(region), index="node"),
            self.get_table_and_convert(self.three_winding_dummy_nodes_query(region), index="node"),
        )
        df = pd.concat(dfs)
        return BusDataSchema(df)

    def loads_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*", "ssh_graph": self.config.ssh_graph}
        return self.template_to_query(templates.LOADS_QUERY, substitutes)

    async def loads(self, region: Optional[str] = None) -> LoadsSchema:
        """Query load data.

        Args:
           region: regexp that limits to region
        """
        query = self.loads_query(region)
        df = await self.get_table_and_convert(query, index="mrid")
        return LoadsSchema(df)

    def wind_generating_units_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.WIND_GENERATING_UNITS_QUERY, substitutes)

    async def wind_generating_units(
        self, region: Optional[str] = None
    ) -> WindGeneratingUnitsSchema:
        """Query wind generating units.

        Args:
           region:

        Example:
            >>> from cimsparql.model import get_cim_model
            >>> server_url = "127.0.0.1:7200"
            >>> model = get_cim_model(server_url, "LATEST")
            >>> model.wind_generating_units()

        """
        query = self.wind_generating_units_query(region)
        df = await self.get_table_and_convert(query, index="mrid")
        return WindGeneratingUnitsSchema(df)

    def synchronous_machines_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*", "ssh_graph": self.config.ssh_graph}
        return self.template_to_query(templates.SYNCHRONOUS_MACHINES_QUERY, substitutes)

    async def synchronous_machines(self, region: Optional[str] = None) -> SynchronousMachinesSchema:
        query = self.synchronous_machines_query(region)
        df = await self.get_table_and_convert(query, index="mrid")
        return SynchronousMachinesSchema(df)

    def connections_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.CONNECTIONS_QUERY, substitutes)

    async def connections(self, region: Optional[str] = None) -> ConnectionsSchema:
        """Query connectors

        Args:
           rdf_types: Only cim:breaker and cim:Disconnector allowed
        Returns:

        Example:
           region: Limit to region

        Example:
            >>> from cimsparql.model import get_cim_model
            >>> server_url = "127.0.0.1:7200"
            >>> model = get_cim_model(server_url, "LATEST")
            >>> model.connections()
        """
        query = self.connections_query(region)
        df = await self.get_table_and_convert(query, index="mrid")
        return ConnectionsSchema(df)

    def borders_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.BORDERS_QUERY, substitutes)

    async def borders(self, region: Optional[str] = None) -> BordersSchema:
        """Retrieve ACLineSegments where one terminal is inside and the other is outside the region

        Args:
            region: Inside area
            limit: return first 'limit' number of rows
        """
        query = self.borders_query(region)
        df = await self.get_table_and_convert(query, index="mrid")
        return BordersSchema(df)

    def exchange_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.EXCHANGE_QUERY, substitutes)

    async def exchange(self, region: Optional[str] = None) -> ExchangeSchema:
        """Retrieve ACLineSegments where one terminal is inside and the other is outside the region.

        Args:
            region: Inside area
        """

        if region is None:
            cols = ["name", "node", "status", "p", "market_code"]
            index = pd.Index([], name="mrid")
            df = pd.DataFrame([], columns=cols, index=index)
            return ExchangeSchema(df)
        query = self.exchange_query(region)
        df = await self.get_table_and_convert(query, index="mrid")
        return ExchangeSchema(df)

    def converters_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*", "ssh_graph": self.config.ssh_graph}
        return self.template_to_query(templates.CONVERTERS_QUERY, substitutes)

    async def converters(self, region: Optional[str] = None) -> ConvertersSchema:
        query = self.converters_query(region)
        df = await self.get_table_and_convert(query, index="mrid")
        return ConvertersSchema(df)

    def transformers_connected_to_converter_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(
            templates.TRANSFORMERS_CONNECTED_TO_CONVERTER_QUERY, substitutes
        )

    async def transformers_connected_to_converter(
        self, region: Optional[str] = None
    ) -> TransfConToConverterSchema:
        """Query list of transformer connected at a converter (Voltage source or DC)

        Args:
           region: Limit to region

        """
        query = self.transformers_connected_to_converter_query(region)
        df = await self.get_table_and_convert(query, index="mrid")
        return TransfConToConverterSchema(df)

    def ac_lines_query(self, region: Optional[str] = None, rate: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*", "rate": rate or "Normal@20"}
        return self.template_to_query(templates.AC_LINE_QUERY, substitutes)

    async def ac_lines(
        self, region: Optional[str] = None, rate: Optional[str] = None
    ) -> AcLinesSchema:
        """Query ac line segments

        Args:
           region: Limit to region

        Example:
            >>> from cimsparql.model import get_cim_model
            >>> server_url = "127.0.0.1:7200"
            >>> model = get_cim_model(server_url, "LATEST")
            >>> model.ac_lines()
        """
        query = self.ac_lines_query(region, rate)
        df = await self.get_table_and_convert(query, index="mrid")
        return AcLinesSchema(df)

    def series_compensators_query(
        self, region: Optional[str] = None, rate: Optional[str] = None
    ) -> str:
        substitutes = {"region": region or ".*", "rate": rate or "Normal@20"}
        return self.template_to_query(templates.SERIES_COMPENSATORS_QUERY, substitutes)

    async def series_compensators(
        self, region: Optional[str] = None, rate: Optional[str] = None
    ) -> BranchComponentSchema:
        """Query series compensators

        Args:
           region: Limit to region
        """
        query = self.series_compensators_query(region, rate)
        df = await self.get_table_and_convert(query, index="mrid")
        return BranchComponentSchema(df)

    def transformers_query(self, region: Optional[str] = None, rate: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*", "rate": rate or "Normal@20"}
        return self.template_to_query(templates.TRANSFORMERS_QUERY, substitutes)

    async def transformers(
        self, region: Optional[str] = None, rate: Optional[str] = None
    ) -> TransformersSchema:
        """Query transformer windings.

        Args:
           region: Limit to region
        """
        query = self.transformers_query(region, rate)
        df = await self.get_table_and_convert(query)
        return TransformersSchema(df)

    def two_winding_transformers_query(
        self, region: Optional[str] = None, rate: Optional[str] = None
    ) -> str:
        substitutes = {"region": region or ".*", "rate": rate or "Normal@20"}
        return self.template_to_query(templates.TWO_WINDING_QUERY, substitutes)

    def two_winding_angle_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.TWO_WINDING_ANGLE_QUERY, substitutes)

    async def two_winding_transformers(
        self, region: Optional[str] = None, rate: Optional[str] = None
    ) -> TransformerWindingSchema:
        """Query two-winding transformer.

        Args:
           region: Limit to region

        Example:
            >>> from cimsparql.model import get_cim_model
            >>> server_url = "127.0.0.1:7200"
            >>> model = get_cim_model(server_url, "LATEST")
            >>> model.two_winding_transformers()
        """
        query = self.two_winding_transformers_query(region, rate)
        query_angle = self.two_winding_angle_query(region)
        data = await self.get_table_and_convert(query, index="mrid")
        angle = await self.get_table_and_convert(query_angle, index="mrid")
        if not angle.empty:
            data["angle"] += angle.reindex(index=data.index, fill_value=0.0).squeeze()
        return TransformerWindingSchema(data)

    def three_winding_loss_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.THREE_WINDING_LOSS_QUERY, substitutes)

    def three_winding_transformers_query(
        self, region: Optional[str] = None, rate: Optional[str] = None
    ) -> str:
        substitutes = {"region": region or ".*", "rate": rate or "Normal@20"}
        return self.template_to_query(templates.THREE_WINDING_QUERY, substitutes)

    async def three_winding_transformers(
        self, region: Optional[str] = None, rate: Optional[str] = None
    ) -> TransformerWindingSchema:
        """Query three-winding transformer. Return as three two-winding transformers.

        Example:
            >>> from cimsparql.model import get_cim_model
            >>> server_url = "127.0.0.1:7200"
            >>> model = get_cim_model(server_url, "LATEST")
            >>> model.two_winding_transformers()
        """
        query = self.three_winding_transformers_query(region, rate)
        data = await self.get_table_and_convert(query, index="mrid")
        query_loss = self.three_winding_loss_query(region)
        loss = await self.get_table_and_convert(query_loss, index="mrid")
        df = pd.concat([data.assign(ploss_1=0.0), loss.loc[data.index]], axis=1)
        return TransformerWindingSchema(df)

    @property
    def substation_voltage_level_query(self) -> str:
        return self.template_to_query(templates.SUBSTATION_VOLTAGE_LEVEL_QUERY)

    async def substation_voltage_level(self) -> SubstationVoltageSchema:
        query = self.substation_voltage_level_query
        df = await self.get_table_and_convert(query, index="substation")
        return SubstationVoltageSchema(df)

    @property
    def disconnected_query(self) -> str:
        return self.template_to_query(templates.DISCONNECTED_QUERY)

    async def disconnected(self) -> DisconnectedSchema:
        """Query disconneced status from ssh profile (not available in GraphDB)."""
        df = await self.get_table_and_convert(self.disconnected_query)
        return DisconnectedSchema(df)

    @property
    def powerflow_query(self) -> str:
        return self.template_to_query(templates.POWER_FLOW_QUERY)

    async def powerflow(self) -> PowerFlowSchema:
        """Query powerflow from sv profile (not available in GraphDB)."""
        df = await self.get_table_and_convert(self.powerflow_query, index="mrid")
        return PowerFlowSchema(df)

    @property
    def coordinates_query(self) -> str:
        return self.template_to_query(templates.COORDINATES_QUERY)

    async def coordinates(self) -> CoordinatesSchema:
        return CoordinatesSchema(await self.get_table_and_convert(self.coordinates_query))

    @property
    def st_group_codes_names_query(self) -> str:
        return self.template_to_query(templates.STATION_GROUP_CODE_NAME_QUERY)

    async def station_group_codes_and_names(self) -> StationGroupCodeNameSchema:
        df = await self.get_table_and_convert(self.st_group_codes_names_query)
        return StationGroupCodeNameSchema(df.set_index("station_group"))

    def branch_node_withdraw_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.BRANCH_NODE_WITHDRAW_QUERY, substitutes)

    async def branch_node_withdraw(self, region: Optional[str] = None) -> BranchWithdrawSchema:
        """Query branch flow from sv profile."""
        query = self.branch_node_withdraw_query(region)
        df = await self.get_table_and_convert(query, index="mrid")
        return BranchWithdrawSchema(df)

    def dc_active_flow_query(self, region: Optional[str] = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.DC_ACTIVE_POWER_FLOW_QUERY, substitutes)

    async def dc_active_flow(self, region: Optional[str] = None) -> DcActiveFlowSchema:
        query = self.dc_active_flow_query(region)
        df = await self.get_table_and_convert(query)
        # Unable to group on max within the sparql query so we do it here.
        df = df.iloc[df.groupby("mrid")["p"].idxmax()].set_index("mrid")
        df["p"] *= df["direction"]
        return DcActiveFlowSchema(df.drop(columns="direction"))

    @property
    def regions_query(self) -> str:
        return self.template_to_query(templates.REGIONS_QUERY)

    async def regions(self) -> RegionsSchema:
        """Query regions

        Property:
           regions: List of regions in database

        Example:
            >>> from cimsparql.model import get_cim_model
            >>> server_url = "127.0.0.1:7200"
            >>> model = get_cim_model(server_url, "LATEST")
            >>> model.regions
        """
        # TODO: Probably deprecate this property in the future. But keep for now.
        # We need to find a solution for custom namespaces in queries
        df = await self.get_table_and_convert(self.regions_query, index="mrid")
        return RegionsSchema(df)

    def add_mrid_query(self, rdf_type: Optional[str] = None, graph: Optional[str] = None) -> str:
        substitutes = {"rdf_type": rdf_type or "?rdf_type", "g": graph or "?g"}
        return self.template_to_query(templates.ADD_MRID_QUERY, substitutes)

    def add_mrid(self, rdf_type: Optional[str] = None, graph: Optional[str] = None) -> None:
        """
        Add cim:IdentifiedObject.mRID to all records. It is copied from rdf:about (or rdf:ID) if
        replace is not specified

        Args:
            graph: Name of graph where mrids should be added. Note, mrid is added to all objects
                in the graph.
            rdf_type: RDF type where ID should be added
        """
        self.client.update_query(self.add_mrid_query(rdf_type, graph))


def get_cim_model(
    service_cfg: Optional[ServiceConfig] = None,
    model_cfg: Optional[ModelConfig] = None,
    async_sparql_wrapper: bool = False,
) -> CimModel:
    """Get a CIM Model.


    Args:
        service_cfg: Configurations for the triple store service
        model_cfg: Conifugrations for the CIM mode
        async_sparql_wrapper: If True http calls are made via asynchronous requests.
            If False, the native SparqlWrapper sends requests via urllib
    """
    c = AsyncGraphDBClient(service_cfg) if async_sparql_wrapper else GraphDBClient(service_cfg)
    return CimModel(c, model_cfg)



from __future__ import annotations

import warnings
from contextlib import contextmanager
from datetime import datetime
from decimal import Decimal
from typing import Any, Callable, Dict, Generator, Optional, Union

import pandas as pd

from cimsparql.graphdb import GraphDBClient, ServiceConfig
from cimsparql.templates import TYPE_MAPPER_QUERY

# Type-caster that can be used with pandas. It can be python types, numpy dtypes or string
# value. Examples: float, "int", int, "string"
TYPE_CASTER = Union[Callable[[Any], Any], str]
SPARQL_TYPE = str
COL_NAME = str
PREFIX = str
URI = str

as_type_able = [int, float, "string", "Int64", "Int32", "Int16"]


def to_timedelta(duration: str) -> datetime.timedelta:
    # Pandas only supports days, hours, minites, seconds
    # not year and month which can be part of
    # https://www.w3.org/TR/xmlschema11-2/#duration

    if "Y" in duration:
        raise ValueError("Cimsparql uses pandas to convert duration. Y not supported")

    return pd.to_timedelta(duration)


XSD_TYPE_MAP = {
    # Primitive types (https://www.w3.org/TR/xmlschema11-2/#built-in-primitive-datatypes)
    "boolean": lambda x: x.lower() in {"true", "1"},
    "date": pd.to_datetime,
    "dateTime": pd.to_datetime,
    "decimal": Decimal,
    "double": float,
    "duration": to_timedelta,  # Require ISO 8601 format,
    "float": float,
    "integer": int,
    "time": pd.to_datetime,
}


CIM_TYPE_MAP = {
    "String": "string",
    "Integer": int,
    "Boolean": bool,
    "Float": float,
    "Date": pd.to_datetime,
}

sparql_type_map = {
    "literal": "string",
    "uri": "string",
}


@contextmanager
def enforce_no_limit(client: GraphDBClient) -> Generator[GraphDBClient, None, None]:
    orig_limit = client.service_cfg.limit
    client.service_cfg.limit = None
    try:
        yield client
    finally:
        client.service_cfg.limit = orig_limit


def build_type_map(prefixes: Dict[PREFIX, URI]) -> Dict[SPARQL_TYPE, TYPE_CASTER]:
    short_map = {"xsd": XSD_TYPE_MAP, "cim": CIM_TYPE_MAP}

    type_map = {}
    for namespace, prim_types in short_map.items():
        if namespace not in prefixes:
            continue
        prefix = prefixes[namespace]
        new_map = {f"{prefix}{dtype}": converter for dtype, converter in prim_types.items()}
        type_map.update(new_map)
    return type_map


class TypeMapper:
    def __init__(
        self,
        service_cfg: Optional[ServiceConfig] = None,
        custom_additions: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.client = GraphDBClient(service_cfg)

        self.query = TYPE_MAPPER_QUERY.substitute(self.client.prefixes)
        custom_additions = custom_additions or {}
        self.prim_type_map = build_type_map(self.client.prefixes)
        self.map = sparql_type_map | self.get_map() | self.prim_type_map | custom_additions

    def have_cim_version(self, cim) -> bool:
        return any(cim in val for val in self.map.keys())

    def type_map(self, df: pd.DataFrame) -> Dict[str, Any]:
        return {
            row.sparql_type: self.prim_type_map.get(row.range, "string") for row in df.itertuples()
        }

    def get_map(self) -> Dict[str, Any]:
        """Reads all metadata from the sparql backend & creates a sparql-type -> python type map

        Args:
            client: initialized CimModel

        Returns:
            sparql-type -> python type map

        """

        with enforce_no_limit(self.client) as c:
            res = c.get_table(self.query)
            df = res[0]
        if df.empty:
            return {}
        return self.type_map(df)

    def get_type(self, sparql_type: str, col: str = ""):
        """Gets the python type/function to apply on columns of the sparql_type

        Args:
            sparql_type:
            missing_return: returns the identity-function if python- type/function is not found,
                else returns None

            custom_maps: dictionary on the form {'sparql_data_type': function/datatype} overwrites
                the default types gained from the graphdb. Applies the function/datatype on all
                columns in the DataFrame that are of the sparql_data_type

        Returns:
            python datatype or function to apply on DataFrame columns

        """

        try:
            return self.map[sparql_type]
        except KeyError:
            warnings.warn(
                f"{col}:{sparql_type} not found in the sparql -> python type map. Using 'string'"
            )
        return "string"

    def build_type_caster(
        self, col_map: Dict[COL_NAME, SPARQL_TYPE]
    ) -> Dict[COL_NAME, TYPE_CASTER]:
        """
        Construct a direct mapping from column names to a type caster from the
        """
        return {col: self.get_type(dtype, col) for col, dtype in col_map.items()}

    def map_data_types(
        self, df: pd.DataFrame, col_map: Dict[COL_NAME, SPARQL_TYPE]
    ) -> pd.DataFrame:
        """Maps the dtypes of a DataFrame to the python-corresponding types of the sparql-types from
        the source data

        Args:
            df: DataFrame with columns to be converted

        Returns:
            mapped DataFrame

        """
        if df.empty:
            return df
        type_caster = self.build_type_caster(col_map)
        df = map_base_types(df, type_caster)
        df = map_exceptions(df, type_caster)
        return df


def map_base_types(df: pd.DataFrame, type_map: Dict[COL_NAME, TYPE_CASTER]) -> pd.DataFrame:
    """Maps the datatypes in type_map which can be used with the df.astype function

    Args:
        df:

    Returns:
        mapped DataFrame

    """
    as_type_able_columns = {c for c, datatype in type_map.items() if datatype in as_type_able}
    if not df.empty:
        df = df.astype({column: type_map[column] for column in as_type_able_columns})
    return df


def map_exceptions(df: pd.DataFrame, type_map: Dict[COL_NAME, TYPE_CASTER]) -> pd.DataFrame:
    """Maps the functions/datatypes in type_map which cant be done with the df.astype function

    Args:
        df:

    Returns:
        mapped DataFrame

    """
    ex_columns = {c for c, datatype in type_map.items() if datatype not in as_type_able}
    for column in ex_columns:
        df[column] = df[column].apply(type_map[column])
    return df



import datetime as dt

import pandas as pd
import pandera as pa
from pandera.typing import DataFrame, Index, Series


class JsonSchemaOut(pa.SchemaModel):
    class Config:
        to_format = "json"
        to_format_kwargs = {"orient": "table"}
        coerce = True


class FullModelSchema(JsonSchemaOut):
    model: Series[pd.StringDtype] = pa.Field()
    time: Series[pd.StringDtype] = pa.Field()
    profile: Series[pd.StringDtype] = pa.Field()
    version: Series[pd.StringDtype] = pa.Field()
    description: Series[pd.StringDtype] = pa.Field()

    @pa.dataframe_check
    def unique_time(cls, df: DataFrame) -> bool:
        return len(df["time"].unique()) == 1


class MridResourceSchema(JsonSchemaOut):
    """
    Common class for resources with an mrid as index
    """

    mrid: Index[pd.StringDtype] = pa.Field(unique=True)


class NamedResourceSchema(MridResourceSchema):
    """
    Common class for resources with an mrid and a name
    """

    name: Series[pd.StringDtype] = pa.Field()


class NamedMarketResourceSchema(NamedResourceSchema):
    """
    Common class for named resources with an (optional) associated market_code
    """

    market_code: Series[pd.StringDtype] = pa.Field(nullable=True)


class MarketDatesSchema(NamedResourceSchema):
    activation_date: Series[dt.datetime] = pa.Field()


class BusDataSchema(JsonSchemaOut):
    node: Index[pd.StringDtype] = pa.Field(
        unique=True,
    )
    name: Series[pd.StringDtype] = pa.Field()
    busname: Series[pd.StringDtype] = pa.Field()
    un: Series[float] = pa.Field()
    station: Series[pd.StringDtype] = pa.Field()
    bidzone: Series[pd.StringDtype] = pa.Field(nullable=True)


class LoadsSchema(NamedResourceSchema):
    node: Series[pd.StringDtype] = pa.Field()
    station: Series[pd.StringDtype] = pa.Field()
    bidzone: Series[pd.StringDtype] = pa.Field(nullable=True)
    status: Series[bool] = pa.Field()
    p: Series[float] = pa.Field(nullable=True)
    q: Series[float] = pa.Field(nullable=True)
    station_group: Series[pd.StringDtype] = pa.Field(nullable=True)


class WindGeneratingUnitsSchema(NamedMarketResourceSchema):
    station_group: Series[pd.StringDtype] = pa.Field(nullable=True)
    min_p: Series[float] = pa.Field()
    max_p: Series[float] = pa.Field()
    plant_mrid: Series[pd.StringDtype] = pa.Field(nullable=True)


class SynchronousMachinesSchema(NamedMarketResourceSchema):
    allocationmax: Series[float] = pa.Field(nullable=True)
    node: Series[pd.StringDtype] = pa.Field()
    status: Series[bool] = pa.Field()
    station_group: Series[pd.StringDtype] = pa.Field(nullable=True)
    station_group_name: Series[pd.StringDtype] = pa.Field(nullable=True)
    station: Series[pd.StringDtype] = pa.Field()
    maxP: Series[float] = pa.Field(nullable=True)
    minP: Series[float] = pa.Field(nullable=True)
    MO: Series[float] = pa.Field(nullable=True)
    bidzone: Series[pd.StringDtype] = pa.Field(nullable=True)
    sn: Series[float] = pa.Field()
    p: Series[float] = pa.Field(nullable=True)
    q: Series[float] = pa.Field(nullable=True)


class ConnectionsSchema(MridResourceSchema):
    t_mrid_1: Series[pd.StringDtype] = pa.Field()
    t_mrid_2: Series[pd.StringDtype] = pa.Field()


class BordersSchema(NamedMarketResourceSchema):
    area_1: Series[pd.StringDtype] = pa.Field()
    area_2: Series[pd.StringDtype] = pa.Field()
    t_mrid_1: Series[pd.StringDtype] = pa.Field()
    t_mrid_2: Series[pd.StringDtype] = pa.Field()


class ExchangeSchema(NamedMarketResourceSchema):
    node: Series[pd.StringDtype] = pa.Field()
    status: Series[bool] = pa.Field()
    p: Series[float] = pa.Field()


class ConvertersSchema(NamedResourceSchema):
    alias: Series[pd.StringDtype] = pa.Field(nullable=True)
    station: Series[pd.StringDtype] = pa.Field()
    status: Series[bool] = pa.Field()
    node: Series[pd.StringDtype] = pa.Field()


class TransfConToConverterSchema(NamedResourceSchema):
    t_mrid: Series[pd.StringDtype] = pa.Field()
    p_mrid: Series[pd.StringDtype] = pa.Field()


class CoordinatesSchema(JsonSchemaOut):
    mrid: Series[pd.StringDtype] = pa.Field()
    x: Series[pd.StringDtype] = pa.Field()
    y: Series[pd.StringDtype] = pa.Field()
    epsg: Series[pd.CategoricalDtype] = pa.Field()
    rdf_type: Series[pd.CategoricalDtype] = pa.Field()


class BranchComponentSchema(NamedResourceSchema):
    bidzone_1: Series[pd.StringDtype] = pa.Field(nullable=True)
    bidzone_2: Series[pd.StringDtype] = pa.Field(nullable=True)
    node_1: Series[pd.StringDtype] = pa.Field()
    node_2: Series[pd.StringDtype] = pa.Field()
    ploss_1: Series[float] = pa.Field(nullable=True)
    ploss_2: Series[float] = pa.Field(nullable=True)
    r: Series[float] = pa.Field()
    rate: Series[float] = pa.Field(nullable=True)
    status: Series[bool] = pa.Field()
    un: Series[float] = pa.Field()
    x: Series[float] = pa.Field()


class ShuntComponentSchema(BranchComponentSchema):
    b: Series[float] = pa.Field()
    g: Series[float] = pa.Field()


class AcLinesSchema(ShuntComponentSchema):
    length: Series[float] = pa.Field()
    g: Series[float] = pa.Field(nullable=True)


class TransformersSchema(JsonSchemaOut):
    name: Series[pd.StringDtype] = pa.Field()
    p_mrid: Series[pd.StringDtype] = pa.Field()
    w_mrid: Series[pd.StringDtype] = pa.Field()
    endNumber: Series[int] = pa.Field()
    un: Series[float] = pa.Field()
    t_mrid: Series[pd.StringDtype] = pa.Field()
    r: Series[float] = pa.Field()
    x: Series[float] = pa.Field()
    rate: Series[float] = pa.Field(nullable=True)


class TransformerWindingSchema(ShuntComponentSchema):
    angle: Series[float] = pa.Field()
    ratio: Series[float] = pa.Field()


class SubstationVoltageSchema(JsonSchemaOut):
    substation: Index[pd.StringDtype] = pa.Field()
    container: Series[pd.StringDtype] = pa.Field()
    v: Series[float] = pa.Field()


class DisconnectedSchema(JsonSchemaOut):
    mrid: Series[pd.StringDtype] = pa.Field(unique=True)


class PowerFlowSchema(MridResourceSchema):
    p: Series[float] = pa.Field()
    q: Series[float] = pa.Field()
    in_service: Series[bool] = pa.Field()


class BranchWithdrawSchema(MridResourceSchema):
    node: Series[pd.StringDtype] = pa.Field()
    p: Series[float] = pa.Field()
    q: Series[float] = pa.Field()


class DcActiveFlowSchema(MridResourceSchema):
    p: Series[float] = pa.Field()


class RegionsSchema(MridResourceSchema):
    region: Series[pd.StringDtype] = pa.Field()
    short_name: Series[pd.StringDtype] = pa.Field(nullable=True)
    name: Series[pd.StringDtype] = pa.Field()
    alias_name: Series[pd.StringDtype] = pa.Field(nullable=True)
    region_name: Series[pd.StringDtype] = pa.Field(nullable=True)


class StationGroupCodeNameSchema(JsonSchemaOut):
    station_group: Index[pd.StringDtype] = pa.Field(unique=True)
    name: Series[pd.StringDtype] = pa.Field()
    alias_name: Series[pd.StringDtype] = pa.Field(nullable=True)



"""Functions used to configure GraphDB client

Will handle authenticated instances of GraphDB where user and password is given in environment
variables ("GRAPHDB_USER" & "GRAPHDB_USER_PASSWD").

"""
import os
from typing import Dict, List, Optional

import requests


def service(
    repo: Optional[str] = None,
    server: str = "127.0.0.1:7200",
    protocol: str = "https",
    path: str = "",
) -> str:
    """Returns service url for GraphdDBClient

    Args:
       repo: Repo on server
       server: server ip/name
       protocol: http or https
    """
    url = f"{protocol}://{server}/{path}repositories"
    if repo:
        url += f"/{repo}"
    return url


def service_blazegraph(server: str, repo: str, protocol: str = "https"):
    return f"{protocol}://{server}/{repo}/sparql"


class GraphDbConfig:
    def __init__(
        self,
        server: str = "127.0.0.1:7200",
        protocol: str = "https",
        auth: requests.auth.AuthBase = None,
    ) -> None:
        """Get repo configuration from GraphDB

        Args:
           server: GraphDB server
           protocol: http or https

        """
        self._service = service(None, server, protocol)
        if auth is None:
            auth = requests.auth.HTTPBasicAuth(
                os.getenv("GRAPHDB_USER"), os.getenv("GRAPHDB_USER_PASSWD")
            )
        try:
            response = requests.get(
                self._service, headers={"Accept": "application/json"}, auth=auth
            )
            response.raise_for_status()
            self._repos = response.json()["results"]["bindings"]
        except (requests.exceptions.RequestException, KeyError):
            self._repos: List[Dict[str, Dict[str, str]]] = []

    @property
    def repos(self) -> List[str]:
        """List of available repos on GraphDB server."""
        if self._repos:
            return [repo["id"]["value"] for repo in self._repos]
        return []



"""Graphdb CIM sparql client"""
import json
import os
from dataclasses import dataclass, field
from enum import auto
from pathlib import Path
from typing import Dict, List, Optional, Tuple, TypedDict, Union

import httpx
import pandas as pd
import requests
from SPARQLWrapper import JSON, POST, SPARQLWrapper
from strenum import StrEnum

from cimsparql.async_sparql_wrapper import AsyncSparqlWrapper
from cimsparql.url import service, service_blazegraph


class SparqlResult(TypedDict):
    cols: List[str]
    data: List[Dict[str, Dict[str, str]]]
    out: List[Dict[str, str]]


def data_row(cols: List[str], rows: List[Dict[str, Dict[str, str]]]) -> Dict[str, str]:
    """Get a sample row for extraction of data types

    Args:
       cols: queried columns (optional might return None)
       rows: 'results'→'bindings' from SPARQLWrapper

    Returns:
       samples result of all columns in query
    """
    full_row = {}
    for row in rows:
        if set(cols).difference(full_row):
            full_row.update(row)
        else:
            break
    return full_row


class RestApi(StrEnum):
    RDF4J = auto()
    BLAZEGRAPH = auto()
    DIRECT_SPARQL_ENDPOINT = auto()


def parse_namespaces_rdf4j(response: requests.Response) -> Dict[str, str]:
    prefixes = {}
    for line in response.text.split()[1:]:
        prefix, uri = line.split(",")
        prefixes[prefix] = uri
    return prefixes


@dataclass
class ServiceConfig:
    repo: str = field(default=os.getenv("GRAPHDB_REPO", "LATEST"))
    protocol: str = "https"
    server: str = field(default=os.getenv("GRAPHDB_SERVER", "127.0.0.1:7200"))
    path: str = ""
    user: str = field(default=os.getenv("GRAPHDB_USER"))
    passwd: str = field(default=os.getenv("GRAPHDB_USER_PASSWD"))
    rest_api: RestApi = field(default=os.getenv("SPARQL_REST_API", RestApi.RDF4J))

    # Parameters for rest api
    # https://rdf4j.org/documentation/reference/rest-api/
    distinct: bool = False
    infer: bool = False
    limit: Optional[int] = None
    offset: Optional[int] = None
    timeout: Optional[int] = None

    def __post_init__(self):
        if self.rest_api not in RestApi:
            raise ValueError(f"rest_api must be one of {RestApi}")

    @property
    def url(self) -> str:
        if self.rest_api == RestApi.BLAZEGRAPH:
            return service_blazegraph(self.server, self.repo, self.protocol)
        elif self.rest_api == RestApi.DIRECT_SPARQL_ENDPOINT:
            return self.server
        return service(self.repo, self.server, self.protocol, self.path)

    @property
    def parameters(self) -> Dict[str, Union[bool, int]]:
        return {
            "distinct": self.distinct,
            "infer": self.infer,
            "limit": self.limit,
            "offset": self.offset,
        }


# Available formats from RDF4J API
# https://rdf4j.org/documentation/reference/rest-api/
MIME_TYPE_RDF_FORMATS = {
    "rdf/xml": "application/rdf+xml",
    "n-triples": "text/plain",
    "turtle": "text/turtle",
    "n3": "text/rdf+n3",
    "n-quads": "text/x-nquads",
    "json-ld": "application/ld+json",
    "rdf/json": "application/rdf+json",
    "trix": "application/trix",
    "trig": "application/x-trig",
    "rdf4J binary rdf": "application/x-binary-rdf",
}

UPLOAD_END_POINT = {RestApi.RDF4J: "/statements", RestApi.BLAZEGRAPH: ""}


def require_rdf4j(f):
    def wrapper(*args):
        self = args[0]
        if self.service_cfg.rest_api != RestApi.RDF4J:
            raise NotImplementedError("Function only implemented for RDF4J")
        return f(*args)

    return wrapper


class GraphDBClient:
    def __init__(self, service_cfg: Optional[ServiceConfig] = None) -> None:
        """GraphDB client

        Args:
           service: string with url to graphdb repository. See cimsparql.url.service
           mapper: GraphDBClient with the mapper (Default to self).
           infer: deduce further knowledge based on existing RDF data and a formal set of
           sameas: map same concepts from two or more datasets
        """
        self.service_cfg = service_cfg or ServiceConfig()
        self.sparql = SPARQLWrapper(self.service_cfg.url)
        self._init_sparql_wrapper()
        self._prefixes = None

    def _init_sparql_wrapper(self):
        self.sparql.setReturnFormat(JSON)
        self.sparql.setMethod(POST)
        self.sparql.setCredentials(self.service_cfg.user, self.service_cfg.passwd)
        if self.service_cfg.timeout:
            self.sparql.setTimeout(self.service_cfg.timeout)
        self._update_sparql_parameters()

    def _update_sparql_parameters(self):
        for key, value in self.service_cfg.parameters.items():
            if value is not None:
                self.set_parameter(key, str(value))

    def set_parameter(self, key: str, value: str) -> None:
        self.sparql.clearParameter(key)
        self.sparql.addParameter(key, value)

    def set_repo(self, repo: str) -> None:
        self.service_cfg.repo = repo
        self.sparql.endpoint = self.service_cfg.url

    @property
    def prefixes(self) -> Dict[str, str]:
        if self._prefixes is None:
            self._prefixes = self.get_prefixes()
        return self._prefixes

    def update_prefixes(self, pref: Dict[str, str]):
        """
        Update prefixes from a dict
        """
        self.prefixes.update(pref)

    def __str__(self) -> str:
        return f"<GraphDBClient object, service: {self.service_cfg.url}>"

    def _prep_query(self, query: str):
        self.sparql.setQuery(query)
        self._update_sparql_parameters()

    def _process_result(self, results: dict) -> dict:
        cols = results["head"]["vars"]
        data = results["results"]["bindings"]
        out = [{c: row.get(c, {}).get("value") for c in cols} for row in data]
        return {"out": out, "cols": cols, "data": data}

    def _exec_query(self, query: str) -> SparqlResult:
        self._prep_query(query)
        results = self.sparql.queryAndConvert()
        return self._process_result(results)

    def exec_query(self, query: str) -> List[Dict[str, str]]:
        out = self._exec_query(query)
        return out["out"]

    def _convert_query_result_to_df(self, res: SparqlResult) -> Tuple[pd.DataFrame, Dict[str, str]]:
        df = pd.DataFrame(res["out"]) if len(res["out"]) else pd.DataFrame(columns=res["cols"])
        return df, data_row(res["cols"], res["data"])

    def get_table(self, query: str) -> Tuple[pd.DataFrame, Dict[str, str]]:
        """
        Args:
           query: to sparql server
           limit: limit number of resulting rows
        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> query = 'select * where { ?subject ?predicate ?object }'
           >>> gdbc.get_table(query, limit=10)
        """
        res = self._exec_query(query)
        return self._convert_query_result_to_df(res)

    @property
    def empty(self) -> bool:
        """Identify empty GraphDB repo"""
        return self.get_table("select * where {?s ?p ?o} limit 1")[0].empty

    def get_prefixes(self) -> Dict[str, str]:
        prefixes = default_namespaces()

        if self.service_cfg.rest_api in (RestApi.BLAZEGRAPH, RestApi.DIRECT_SPARQL_ENDPOINT):
            # These APis does not expose prefixes. Custom prefixes must be added
            # via `update_prefixes`. By default we load a pre-defined set of prefixes
            return prefixes

        auth = requests.auth.HTTPBasicAuth(self.service_cfg.user, self.service_cfg.passwd)
        response = requests.get(self.service_cfg.url + "/namespaces", auth=auth)
        if response.ok:
            prefixes.update(parse_namespaces_rdf4j(response))
            return prefixes
        msg = (
            "Could not fetch namespaces and prefixes from graphdb "
            "Verify that user and password are correctly set in the "
            "GRAPHDB_USER and GRAPHDB_USER_PASSWD environment variable"
        )
        raise RuntimeError(f"{msg} Status code: {response.status_code} Reason: {response.reason}")

    def delete_repo(self) -> None:
        endpoint = delete_repo_endpoint(self.service_cfg)
        response = requests.delete(endpoint)
        response.raise_for_status()

    def upload_rdf(
        self, fname: Path, rdf_format: str, params: Optional[Dict[str, str]] = None
    ) -> None:
        """
        Upload data in RDF format to a srevice

        Args
            fname: Filename to the RDF file to upload
            rdf_format: RDF file type (e.g. rdf/xml, rdf/json etc. Consult the RDF4J rest api
                doc for a complete list of available options)
            params: Additional parameters passed to the post request
        """
        with open(fname, "rb") as infile:
            xml_content = infile.read()

        response = requests.post(
            self.service_cfg.url + UPLOAD_END_POINT[self.service_cfg.rest_api],
            data=xml_content,
            params=params,
            headers={"Content-Type": MIME_TYPE_RDF_FORMATS[rdf_format]},
        )
        response.raise_for_status()

    def update_query(self, query: str) -> None:
        """
        Function that passes a query via a post API call
        """
        auth = requests.auth.HTTPBasicAuth(self.service_cfg.user, self.service_cfg.passwd)
        response = requests.post(
            self.service_cfg.url + UPLOAD_END_POINT[self.service_cfg.rest_api],
            data={"update": query},
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            auth=auth,
        )
        response.raise_for_status()

    @require_rdf4j
    def set_namespace(self, prefix: str, value: str) -> None:
        auth = requests.auth.HTTPBasicAuth(self.service_cfg.user, self.service_cfg.passwd)
        response = requests.put(
            self.service_cfg.url + f"/namespaces/{prefix}",
            data=value,
            headers={"Content-Type": "text/plain"},
            auth=auth,
        )
        response.raise_for_status()

    @require_rdf4j
    def get_namespace(self, prefix: str) -> str:
        auth = requests.auth.HTTPBasicAuth(self.service_cfg.user, self.service_cfg.passwd)
        response = requests.get(self.service_cfg.url + f"/namespaces/{prefix}", auth=auth)
        response.raise_for_status()
        return response.text


@dataclass
class RepoInfo:
    uri: str
    repo_id: str
    title: str
    readable: bool
    writable: bool


async def repos(service_cfg: Optional[ServiceConfig] = None) -> List[RepoInfo]:
    """
    List available repositories
    """

    service_cfg = service_cfg or ServiceConfig()

    auth = None
    if service_cfg.user and service_cfg.passwd:
        auth = httpx.BasicAuth(service_cfg.user, service_cfg.passwd)

    url = f"{service_cfg.protocol}://{service_cfg.server}/repositories"

    async with httpx.AsyncClient() as client:
        response = await client.get(url, auth=auth, headers={"Accept": "application/json"})
    response.raise_for_status()

    def _repo_info(repo):
        uri = repo["uri"]["value"]
        repo_id = repo["id"]["value"]
        title = repo["title"]["value"]
        readable = repo["readable"]["value"] == "true"
        writable = repo["writable"]["value"] == "true"
        return RepoInfo(uri, repo_id, title, readable, writable)

    return [_repo_info(repo) for repo in response.json()["results"]["bindings"]]


def new_repo(
    server: str, repo: str, config: bytes, allow_exist: bool = True, protocol: str = "http"
) -> GraphDBClient:
    """
    Initiialzie a new repository

    Args:
        server: URL to service
        repo: Name of repo
        config: Bytes representaiton of a Turtle config file
        allow_exist: If True, a client pointing to the existing repo is returned.
            Otherwise an error is raised if a repository with the same name already exists
        protocol: Default https
    """
    ignored_errors = {409} if allow_exist else set()
    url = service(repo, server, protocol)
    response = requests.put(url, data=config, headers={"Content-Type": "text/turtle"})
    if response.status_code not in ignored_errors:
        response.raise_for_status()

    return GraphDBClient(ServiceConfig(repo, protocol, server))


def new_repo_blazegraph(url: str, repo: str, protocol: str = "https") -> GraphDBClient:
    template = confpath() / "blazegraph_repo_config.xml"
    config = config_bytes_from_template(template, {"repo": repo})

    response = requests.post(
        f"{protocol}://{url}", data=config, headers={"Content-type": "application/xml"}
    )
    response.raise_for_status()
    client = GraphDBClient(ServiceConfig(repo, protocol, url, rest_api=RestApi.BLAZEGRAPH))
    return client


def config_bytes_from_template(
    template: Path, params: Dict[str, str], encoding: str = "utf8"
) -> bytes:
    """
    Replace value in template file with items in params

    Args:
        template: Path to template file
        params: Dict with key-value pairs where key must be enclode by double curly braces
            in the template file ({{}}). {{key}} will be replaced by value
    """
    with open(template, "rb") as infile:
        data = infile.read()

    for param, value in params.items():
        enc_templ = f"{{{{{param}}}}}".encode(encoding)
        enc_value = value.encode(encoding)
        data = data.replace(enc_templ, enc_value)
    return data


def confpath() -> Path:
    return Path(__file__).parent.parent / "pkg_data"


def default_namespaces() -> Dict[str, str]:
    with open(confpath() / "namespaces.json") as infile:
        return json.load(infile)


def delete_repo_endpoint(config: ServiceConfig) -> str:
    if config.rest_api == RestApi.BLAZEGRAPH:
        # Remove /sparql at the end
        return config.url.rpartition("/")[0]
    return config.url


class AsyncGraphDBClient(GraphDBClient):
    def __init__(self, service_cfg: Optional[ServiceConfig] = None) -> None:
        super().__init__(service_cfg)
        self.sparql = AsyncSparqlWrapper(self.service_cfg.url)
        self._init_sparql_wrapper()

    async def _exec_query(self, query: str) -> SparqlResult:
        self._prep_query(query)
        results = await self.sparql.queryAndConvert()
        return self._process_result(results)

    async def exec_query(self, query: str) -> List[Dict[str, str]]:
        out = await self._exec_query(query)
        return out["out"]

    async def get_table(self, query: str) -> Tuple[pd.DataFrame, Dict[str, str]]:
        """
        Args:
           query: to sparql server
           limit: limit number of resulting rows
        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> query = 'select * where { ?subject ?predicate ?object }'
           >>> gdbc.get_table(query, limit=10)
        """
        res = await self._exec_query(query)
        return self._convert_query_result_to_df(res)


def make_async(client: GraphDBClient) -> AsyncGraphDBClient:
    """
    Convenience function that creates a new async graph db client from an existing client
    """
    return AsyncGraphDBClient(client.service_cfg)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cimsparql']

package_data = \
{'': ['*'], 'cimsparql': ['sparql/*']}

install_requires = \
['SPARQLWrapper', 'StrEnum', 'httpx', 'pandas', 'pandera', 'requests']

setup_kwargs = {
    'name': 'cimsparql',
    'version': '2.3.2',
    'description': 'CIM query utilities',
    'long_description': '[![PyPI version](https://img.shields.io/pypi/v/cimsparql)](https://pypi.org/project/cimsparql/)\n[![Python Versions](https://img.shields.io/pypi/pyversions/cimsparql)](https://pypi.org/project/cimsparql/)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![](https://github.com/statnett/data_cache/workflows/Tests/badge.svg)](https://github.com/statnett/cimsparql/actions?query=workflow%3ATests)\n[![codecov](https://codecov.io/gh/statnett/cimsparql/branch/master/graph/badge.svg)](https://codecov.io/gh/statnett/cimsparql)\n\n# CIMSPARQL Query CIM data using sparql\n\nThis Python package provides functionality for reading cim data from\ntripple stores such as GraphDB, BlazeGraph or Rdf4j into Python memory\nas pandas dataframes.\n\nThe package provides a set of predefined functions/queries to load CIM\ndata such as generator, demand or branch data, though the user can\neasiliy define their own queries.\n\n## Usage\n\n### Load data using predefined functions/queries\n\n```python\n>>> from cimsparql.graphdb import ServiceConfig\n>>> from cimsparql.model import get_cim_model\n>>> model = get_cim_model(ServiceConfig(limit=3))\n>>> ac_lines = model.ac_lines()\n>>> print(ac_lines[[\'name\', \'x\', \'r\', \'bch\']])\n         name       x       r       bch\n0  <branch 1>  1.9900  0.8800  0.000010\n1  <branch 2>  1.9900  0.8800  0.000010\n2  <branch 3>  0.3514  0.1733  0.000198\n```\n\nIn the example above the client will query repo "<repo>" in the default server\n[GraphDB](https://graphdb.ontotext.com) for AC line values.\n\n### Inspect/view predefined queries\n\nSee the sparql templates folder (`cimsparql/sparql`) to the query used.\n\n### Load data using user specified queries\n\n```python\n>>> from string import Template\n>>> query = \'PREFIX cim:<${cim}>\\nPREFIX rdf: <${rdf}>\\nSELECT ?mrid where {?mrid rdf:type cim:ACLineSegment}\'\n>>> query_result = model.get_table_and_convert(model.template_to_query(Template(query)))\n>>> print(query_result)\n```\n\n### List of available repos at the server\n\n```python\n>>> from cimsparql.url import GraphDbConfig\n>>> print(GraphDbConfig().repos)\n```\n\n### Prefix and namespace\n\nAvailable namespace for current graphdb client (`gdbc` in the examples above),\nwhich can be used in queries (such as `rdf` and `cim`) can by found by\n\n```python\n>>> print(model.prefixes)\n{\'wgs\': \'http://www.w3.org/2003/01/geo/wgs84_pos#\',\n \'rdf\': \'http://www.w3.org/1999/02/22-rdf-syntax-ns#\',\n \'owl\': \'http://www.w3.org/2002/07/owl#\',\n \'cim\': \'http://iec.ch/TC57/2010/CIM-schema-cim15#\',\n \'gn\': \'http://www.geonames.org/ontology#\',\n \'xsd\': \'http://www.w3.org/2001/XMLSchema#\',\n \'rdfs\': \'http://www.w3.org/2000/01/rdf-schema#\',\n \'SN\': \'http://www.statnett.no/CIM-schema-cim15-extension#\',\n \'ALG\': \'http://www.alstom.com/grid/CIM-schema-cim15-extension#\'}\n```\n\n### Running Tests Against Docker Databases\n\nTests can be run agains RDF4J and/or BlazeGraph databases if a container with the correct images are available.\n\n```\ndocker pull eclipse/rdf4j-workbench\ndocker pull openkbs/blazegraph\n```\n\nLaunch one or both containers and specify the following environment variables\n\n```\nRDF4J_URL = "localhost:8080/rdf4j-server"\nBLAZEGRAPH_URL = "localhost:9999/blazegraph/namespace\n```\n**Note 1**: The port numbers may differ depending on your local Docker configurations.\n**Note 2**: You don\'t *have* to install RDF4J or BlazeGraph. Tests requiring these will be skipped in case\nthey are not available. They will in any case be run in the CI pipeline on GitHub (where both always are available).\n\n### Ontology (for developers)\n\nOntologies for the CIM model can be found at (ENTSOE\'s webpages)[https://www.entsoe.eu/digital/common-information-model/cim-for-grid-models-exchange/].\nFor convenience and testing purposes the ontology are located under `tests/data/ontology`. CIM models used for testing purposes in Cimsparql should\nbe stored in N-quads format. In case you have a model in XML format it can be converted to N-quads by launching a DB (for example RDF4J) and upload\nall the XML files and the ontology.\n\nExecute\n\n```sparql\nPREFIX cims: <http://iec.ch/TC57/1999/rdf-schema-extensions-19990926#>\n\nDELETE {?s ?p ?o}\nINSERT {?s ?p ?o_cast} WHERE {\n  ?s ?p ?o .\n  ?p cims:dataType ?_dtype .\n  ?_dtype cims:stereotype ?stereotype .\n  BIND(IF(?stereotype = "Primitive",\n    URI(concat("http://www.w3.org/2001/XMLSchema#", lcase(strafter(str(?_dtype), "#")))),\n    ?_dtype) as ?dtype)\n  BIND(STRDT(?o, ?dtype) as ?o_cast)\n}\n```\nand export as N-quads.\n\n### Test models\n\n1. *micro_t1_nl*: `MicroGrid/Type1_T1/CGMES_v2.4.15_MicroGridTestConfiguration_T1_NL_Complete_v2`\n\n\n### Rest APIs\n\nCimSparql mainly uses `SparqlWrapper` to communicate with the databases. However, there are certain operations which are performed\ndirectly via REST calls. Since there are small differences between different APIs you may have to specify which API you are using.\nThis can be done when initializing the `ServiceCfg` class or by specifying the `SPARQL_REST_API` environment variable. Currently,\n`RDF4J` and `blazegraph` is supported (if not given `RDF4J` is default).\n\n```bash\nexport SPARQL_REST_API=RDF4J  # To use RDF4J\nexport SPARQL_REST_API=BLAZEGRAPH  # To use BlazeGraph\n```\n',
    'author': 'Statnett Datascience',
    'author_email': 'Datascience.Drift@Statnett.no',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/statnett/cimsparql.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
